@@ -54,9 +54,9 @@ PYLS_BLACK_REQVER = '>=0.4.6'
 PYLS_SPYDER_REQVER = '>=0.3.2'
 PYXDG_REQVER = '>=0.26'
 PYZMQ_REQVER = '>=17'
-QDARKSTYLE_REQVER = '>=2.8'
+QDARKSTYLE_REQVER = '>=2.8;<3.0'
 QTAWESOME_REQVER = '>=0.5.7'
-QTCONSOLE_REQVER = '>=5.0.1'
+QTCONSOLE_REQVER = '>=5.0.3'
 QTPY_REQVER = '>=1.5.0'
 RTREE_REQVER = '>=0.8.3'
 SETUPTOOLS_REQVER = '>=39.0.0'
