@@ -54,6 +54,13 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
         self.extra_selections_dict = {}
         self._restore_selection_pos = None
 
+        # Trailing newlines/spaces trimming
+        self.remove_trailing_spaces = False
+        self.remove_trailing_newlines = False
+
+        # Add a new line when saving
+        self.add_newline = False
+
         # Code snippets
         self.code_snippets = True
 
@@ -99,8 +106,6 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
         # Useful to avoid recomputing while scrolling.
         self.current_cell = None
 
-        self._cell_list = []
-
         def reset_current_cell():
             self.current_cell = None
 
@@ -118,6 +123,15 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
         self.tab_stop_width_spaces = tab_stop_width_spaces
         self.update_tab_stop_width_spaces()
 
+    def set_remove_trailing_spaces(self, flag):
+        self.remove_trailing_spaces = flag
+
+    def set_add_newline(self, add_newline):
+        self.add_newline = add_newline
+
+    def set_remove_trailing_newlines(self, flag):
+        self.remove_trailing_newlines = flag
+
     def update_tab_stop_width_spaces(self):
         self.setTabStopWidth(self.fontMetrics().width(
                              ' ' * self.tab_stop_width_spaces))
@@ -157,7 +171,7 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
         Also assign draw orders to leave current_cell and current_line
         in the backgrund (and avoid them to cover other decorations)
 
-        NOTE: This will remove previous decorations added to  the same key.
+        NOTE: This will remove previous decorations added to the same key.
 
         Args:
             key (str) name of the extra selections group.
@@ -170,6 +184,7 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
 
         for selection in extra_selections:
             selection.draw_order = draw_order
+            selection.kind = key
 
         self.clear_extra_selections(key)
         self.extra_selections_dict[key] = extra_selections
@@ -193,7 +208,7 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
             key (str) name of the extra selections group.
         """
         for decoration in self.extra_selections_dict.get(key, []):
-            self.decorations.remove(decoration, update=False)
+            self.decorations.remove(decoration)
         self.extra_selections_dict[key] = []
         self.update()
 
@@ -201,6 +216,31 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
         """Emit changed signal"""
         self.modificationChanged.emit(self.document().isModified())
 
+    def get_visible_block_numbers(self):
+        """Get the first and last visible block numbers."""
+        first = self.firstVisibleBlock().blockNumber()
+        bottom_right = QPoint(self.viewport().width() - 1,
+                              self.viewport().height() - 1)
+        last = self.cursorForPosition(bottom_right).blockNumber()
+        return (first, last)
+
+    def get_buffer_block_numbers(self):
+        """
+        Get the first and last block numbers of a region that covers
+        the visible one plus a buffer of half that region above and
+        below to make more fluid certain operations.
+        """
+        first_visible, last_visible = self.get_visible_block_numbers()
+        buffer_height = round((last_visible - first_visible) / 2)
+
+        first = first_visible - buffer_height
+        first = 0 if first < 0 else first
+
+        last = last_visible + buffer_height
+        last = self.blockCount() if last > self.blockCount() else last
+
+        return (first, last)
+
     #------Highlight current line
     def highlight_current_line(self):
         """Highlight current line"""
@@ -219,11 +259,10 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
     #------Highlight current cell
     def highlight_current_cell(self):
         """Highlight current cell"""
-        if not self.has_cell_separators or \
-          not self.highlight_current_cell_enabled:
+        if (not self.has_cell_separators or
+                not self.highlight_current_cell_enabled):
             return
-        cursor, whole_file_selected =\
-            self.select_current_cell()
+        cursor, whole_file_selected = self.select_current_cell()
         selection = TextDecoration(cursor)
         selection.format.setProperty(QTextFormat.FullWidthSelection,
                                      to_qvariant(True))
@@ -239,46 +278,117 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
         """Unhighlight current cell"""
         self.clear_extra_selections('current_cell')
 
-    #------Brace matching
-    def find_brace_match(self, position, brace, forward):
+    def in_comment(self, cursor=None, position=None):
+        """Returns True if the given position is inside a comment.
+
+        Trivial default implementation. To be overridden by subclass.
+        This function is used to define the default behaviour of
+        self.find_brace_match.
+        """
+        return False
+
+    def in_string(self, cursor=None, position=None):
+        """Returns True if the given position is inside a string.
+
+        Trivial default implementation. To be overridden by subclass.
+        This function is used to define the default behaviour of
+        self.find_brace_match.
+        """
+        return False
+
+    def find_brace_match(self, position, brace, forward,
+                         ignore_brace=None, stop=None):
+        """Returns position of matching brace.
+
+        Parameters
+        ----------
+        position : int
+            The position of the brace to be matched.
+        brace : {'[', ']', '(', ')', '{', '}'}
+            The brace character to be matched.
+            [ <-> ], ( <-> ), { <-> }
+        forward : boolean
+            Whether to search forwards or backwards for a match.
+        ignore_brace : callable taking int returning boolean, optional
+            Whether to ignore a brace (as function of position).
+        stop : callable taking int returning boolean, optional
+            Whether to stop the search early (as function of position).
+
+        If both *ignore_brace* and *stop* are None, then brace matching
+        is handled differently depending on whether *position* is
+        inside a string, comment or regular code. If in regular code,
+        then any braces inside strings and comments are ignored. If in a
+        string/comment, then only braces in the same string/comment are
+        considered potential matches. The functions self.in_comment and
+        self.in_string are used to determine string/comment/code status
+        of characters in this case.
+
+        If exactly one of *ignore_brace* and *stop* is None, then it is
+        replaced by a function returning False for every position. I.e.:
+            lambda pos: False
+
+        Returns
+        -------
+        The position of the matching brace. If no matching brace
+        exists, then None is returned.
+        """
+
+        if ignore_brace is None and stop is None:
+            if self.in_string(position=position):
+                # Only search inside the current string
+                def stop(pos):
+                    return not self.in_string(position=pos)
+            elif self.in_comment(position=position):
+                # Only search inside the current comment
+                def stop(pos):
+                    return not self.in_comment(position=pos)
+            else:
+                # Ignore braces inside strings and comments
+                def ignore_brace(pos):
+                    return (self.in_string(position=pos) or
+                            self.in_comment(position=pos))
+
+        # Deal with search range and direction
         start_pos, end_pos = self.BRACE_MATCHING_SCOPE
         if forward:
-            bracemap = {'(': ')', '[': ']', '{': '}'}
-            text = self.get_text(position, end_pos)
-            i_start_open = 1
-            i_start_close = 1
+            closing_brace = {'(': ')', '[': ']', '{': '}'}[brace]
+            text = self.get_text(position, end_pos, remove_newlines=False)
         else:
-            bracemap = {')': '(', ']': '[', '}': '{'}
-            text = self.get_text(start_pos, position)
-            i_start_open = len(text)-1
-            i_start_close = len(text)-1
+            # Handle backwards search with the same code as forwards
+            # by reversing the string to be searched.
+            closing_brace = {')': '(', ']': '[', '}': '{'}[brace]
+            text = self.get_text(start_pos, position+1, remove_newlines=False)
+            text = text[-1::-1]  # reverse
 
+        def ind2pos(index):
+            """Computes editor position from search index."""
+            return (position + index) if forward else (position - index)
+
+        # Search starts at the first position after the given one
+        # (which is assumed to contain a brace).
+        i_start_close = 1
+        i_start_open = 1
         while True:
-            if forward:
-                i_close = text.find(bracemap[brace], i_start_close)
-            else:
-                i_close = text.rfind(bracemap[brace], 0, i_start_close+1)
-            if i_close > -1:
-                if forward:
-                    i_start_close = i_close+1
+            i_close = text.find(closing_brace, i_start_close)
+            i_start_close = i_close+1  # next potential start
+            if i_close == -1:
+                return  # no matching brace exists
+            elif ignore_brace is None or not ignore_brace(ind2pos(i_close)):
+                while True:
                     i_open = text.find(brace, i_start_open, i_close)
-                else:
-                    i_start_close = i_close-1
-                    i_open = text.rfind(brace, i_close, i_start_open+1)
-                if i_open > -1:
-                    if forward:
-                        i_start_open = i_open+1
-                    else:
-                        i_start_open = i_open-1
-                else:
-                    # found matching brace
-                    if forward:
-                        return position+i_close
-                    else:
-                        return position-(len(text)-i_close)
-            else:
-                # no matching brace
-                return
+                    i_start_open = i_open+1  # next potential start
+                    if i_open == -1:
+                        # found matching brace, but should we have
+                        # stopped before this point?
+                        if stop is not None:
+                            # There's room for optimization here...
+                            for i in range(1, i_close+1):
+                                if stop(ind2pos(i)):
+                                    return
+                        return ind2pos(i_close)
+                    elif (ignore_brace is None or
+                          not ignore_brace(ind2pos(i_open))):
+                        break  # must find new closing brace
 
     def __highlight(self, positions, color=None, cancel=False):
         if cancel:
@@ -299,23 +409,32 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
         self.update_extra_selections()
 
     def cursor_position_changed(self):
-        """Brace matching"""
+        """Handle brace matching."""
+        # Clear last brace highlight (if any)
         if self.bracepos is not None:
             self.__highlight(self.bracepos, cancel=True)
             self.bracepos = None
+
+        # Get the current cursor position, check if it is at a brace,
+        # and, if so, determine the direction in which to search for able
+        # matching brace.
         cursor = self.textCursor()
         if cursor.position() == 0:
             return
         cursor.movePosition(QTextCursor.PreviousCharacter,
                             QTextCursor.KeepAnchor)
         text = to_text_string(cursor.selectedText())
-        pos1 = cursor.position()
         if text in (')', ']', '}'):
-            pos2 = self.find_brace_match(pos1, text, forward=False)
+            forward = False
         elif text in ('(', '[', '{'):
-            pos2 = self.find_brace_match(pos1, text, forward=True)
+            forward = True
         else:
             return
+
+        pos1 = cursor.position()
+        pos2 = self.find_brace_match(pos1, text, forward=forward)
+
+        # Set a new brace highlight
         if pos2 is not None:
             self.bracepos = (pos1, pos2)
             self.__highlight(self.bracepos, color=self.matched_p_color)
@@ -380,23 +499,10 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
             super(TextEditBaseWidget, self).keyPressEvent(event)
 
     #------Text: get, set, ...
-    def add_to_cell_list(self, oedata):
-        """Add new cell to cell list."""
-        self._cell_list.append(oedata)
-
     def get_cell_list(self):
         """Get all cells."""
-        # Filter out old cells
-
-        def good(oedata):
-            return oedata.is_valid() and oedata.def_type == oedata.CELL
-
-        self._cell_list = [
-            oedata for oedata in self._cell_list if good(oedata)]
-
-        return sorted(
-            {oedata.get_block_number(): oedata
-             for oedata in self._cell_list}.items())
+        # Reimplemented in childrens
+        return []
 
     def get_selection_as_executable_code(self, cursor=None):
         """Return selected text as a processed text,
@@ -632,19 +738,21 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
         cursor.insertText(text)
         cursor.endEditBlock()
 
+        self.document_did_change()
+
     def duplicate_line_down(self):
         """
         Copy current line or selected text and paste the duplicated text
         *after* the current line or selected text.
         """
-        self.__duplicate_line_or_selection(after_current_line=True)
+        self.__duplicate_line_or_selection(after_current_line=False)
 
     def duplicate_line_up(self):
         """
         Copy current line or selected text and paste the duplicated text
         *before* the current line or selected text.
         """
-        self.__duplicate_line_or_selection(after_current_line=False)
+        self.__duplicate_line_or_selection(after_current_line=True)
 
     def __move_line_or_selection(self, after_current_line=True):
         """Move current line or selected text"""
@@ -654,7 +762,6 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
         last_line = False
 
         # ------ Select text
-
         # Get selection start location
         cursor.setPosition(start_pos)
         cursor.movePosition(QTextCursor.StartOfBlock)
@@ -673,7 +780,6 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
                 last_line = True
 
         # ------ Stop if at document boundary
-
         cursor.setPosition(start_pos)
         if cursor.atStart() and not after_current_line:
             # Stop if selection is already at top of the file while moving up
@@ -691,11 +797,9 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
             return
 
         # ------ Move text
-
         sel_text = to_text_string(cursor.selectedText())
         cursor.removeSelectedText()
 
-
         if after_current_line:
             # Shift selection down
             text = to_text_string(cursor.block().text())
@@ -725,6 +829,8 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
         self.setTextCursor(cursor)
         self.__restore_selection(start_pos, end_pos)
 
+        self.document_did_change()
+
     def move_line_up(self):
         """Move up current line or selected text"""
         self.__move_line_or_selection(after_current_line=False)
@@ -751,9 +857,10 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
                                 QTextCursor.KeepAnchor)
         self.setTextCursor(cursor)
 
-    def delete_line(self):
+    def delete_line(self, cursor=None):
         """Delete current line."""
-        cursor = self.textCursor()
+        if cursor is None:
+            cursor = self.textCursor()
         if self.has_selected_text():
             self.extend_selection_to_complete_lines()
             start_pos, end_pos = cursor.selectionStart(), cursor.selectionEnd()
@@ -880,19 +987,21 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
             # Handle completions for the internal console
             self.insert_text(text)
         else:
-            if self.code_snippets:
-                self.sig_insert_completion.emit(text)
-            else:
-                self.insert_text(text)
+            self.sig_insert_completion.emit(text)
             self.document_did_change()
 
     def is_completion_widget_visible(self):
         """Return True is completion list widget is visible"""
-        return self.completion_widget.isVisible()
+        try:
+            return self.completion_widget.isVisible()
+        except RuntimeError:
+            # This is to avoid a RuntimeError exception when the widget is
+            # already been deleted. See spyder-ide/spyder#13248.
+            return False
 
-    def hide_completion_widget(self):
+    def hide_completion_widget(self, focus_to_parent=True):
         """Hide completion widget and tooltip."""
-        self.completion_widget.hide()
+        self.completion_widget.hide(focus_to_parent=focus_to_parent)
         QToolTip.hideText()
 
     #------Standard keys
@@ -985,7 +1094,6 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
         """Reimplemented to handle focus"""
         self.focus_changed.emit()
         self.focus_in.emit()
-        self.highlight_current_cell()
         QPlainTextEdit.focusInEvent(self, event)
 
     def focusOutEvent(self, event):
@@ -1009,14 +1117,20 @@ class TextEditBaseWidget(QPlainTextEdit, BaseEditMixin):
                     elif event.delta() > 0:
                         self.zoom_in.emit()
                 return
+
         QPlainTextEdit.wheelEvent(self, event)
-        # Needed to prevent stealing focus to the find widget when scrolling
+
+        # Needed to prevent stealing focus when scrolling.
+        # If the current widget with focus is the CompletionWidget, it means
+        # it's being displayed in the editor, so we need to hide it and give
+        # focus back to the editor. If not, we need to leave the focus in
+        # the widget that currently has it.
         # See spyder-ide/spyder#11502
         current_widget = QApplication.focusWidget()
-        self.hide_completion_widget()
-        self.highlight_current_cell()
-        if current_widget is not None:
-            current_widget.setFocus()
+        if isinstance(current_widget, CompletionWidget):
+            self.hide_completion_widget(focus_to_parent=True)
+        else:
+            self.hide_completion_widget(focus_to_parent=False)
 
     def position_widget_at_cursor(self, widget):
         # Retrieve current screen height
